%baseclass-preinclude "semantics.h"

%lsp-needed

%token TOKEN_PROGRAM
%token TOKEN_BEGIN
%token TOKEN_END
%token TOKEN_INTEGER 
%token TOKEN_BOOLEAN
%token TOKEN_SKIP
%token TOKEN_IF
%token TOKEN_THEN
%token TOKEN_ELSE
%token TOKEN_ENDIF
%token TOKEN_WHILE
%token TOKEN_DO
%token TOKEN_DONE
%token TOKEN_READ
%token TOKEN_WRITE
%token TOKEN_SEMICOLON
%token TOKEN_ASSIGN
%token TOKEN_OPEN
%token TOKEN_CLOSE
%token TOKEN_NUM
%token TOKEN_TRUE
%token TOKEN_FALSE
%token <name> TOKEN_ID

%left TOKEN_OR
%left TOKEN_AND
%left TOKEN_NOT
%left TOKEN_EQ
%left TOKEN_LESS TOKEN_GR
%left TOKEN_ADD TOKEN_SUB
%left TOKEN_MUL TOKEN_DIV TOKEN_MOD

%type <expr_type> expression

%union
{
	std::string *name;
	type *expr_type;
}

%%

start:
    TOKEN_PROGRAM TOKEN_ID declarations TOKEN_BEGIN instructions TOKEN_END
    {
    }
;

declarations:
    // empty
    {
    }
|
    declaration declarations
    {
    }
;

declaration:
    TOKEN_INTEGER TOKEN_ID TOKEN_SEMICOLON
    {
		if(symbol_table.count(*$2) > 0) {
			error("Redeclared variable.");
		}
        symbol_table[*$2] = row(type_integer,d_loc__.first_line);
		delete $2;
    }
|
    TOKEN_BOOLEAN TOKEN_ID TOKEN_SEMICOLON
    {
		if(symbol_table.count(*$2) > 0) {
			error("Redeclared variable.");
		}
        symbol_table[*$2] = row(type_boolean,d_loc__.first_line);
		delete $2;
    }
;

instructions:
    instruction
    {
    }
|
    instruction instructions
    {
    }
;

instruction:
    TOKEN_SKIP TOKEN_SEMICOLON
    {
    }
|
    assignment
    {
    }
|
    read
    {
    }
|
    write
    {
    }
|
    branch
    {
    }
|
    loop
    {
    }
;

assignment:
    TOKEN_ID TOKEN_ASSIGN expression TOKEN_SEMICOLON
    {
        if(symbol_table.count(*$1) < 1) {
			std::stringstream ss;
			ss << "Undeclared variable: " << *$1;
			error(ss.str().c_str());
        }
        if(symbol_table[*$1].var_type != *$3) {
			error("Type error in assignment.");
        }
		delete $1;
		delete $3;
    }
;

read:
    TOKEN_READ TOKEN_OPEN TOKEN_ID TOKEN_CLOSE TOKEN_SEMICOLON
    {
        if(symbol_table.count(*$3) < 1) {
			std::stringstream ss;
			ss << "Undeclared variable: " << *$3;
			error(ss.str().c_str());
        }
		delete $3;
    }
;

write:
    TOKEN_WRITE TOKEN_OPEN expression TOKEN_CLOSE TOKEN_SEMICOLON
    {
		delete $3;
    }
;

branch:
    TOKEN_IF expression TOKEN_THEN instructions TOKEN_ENDIF
    {
		if(*$2 != type_boolean) {
			error("Condition of if-instruction is not boolean.");
		}
		delete $2;
    }
|
    TOKEN_IF expression TOKEN_THEN instructions TOKEN_ELSE instructions TOKEN_ENDIF
    {
		if(*$2 != type_boolean) {
			error("Condition of if-instruction is not boolean.");
		}
		delete $2;
    }
;

loop:
    TOKEN_WHILE expression TOKEN_DO instructions TOKEN_DONE
    {
		if(*$2 != type_boolean) {
			error("Condition of while-loop is not boolean.");
		}
		delete $2;
    }
;

expression:
    TOKEN_NUM
    {
        $$ = new type(type_integer);
    }
|
    TOKEN_TRUE
    {
        $$ = new type(type_boolean);
    }
|
    TOKEN_FALSE
    {
        $$ = new type(type_boolean);
    }
|
    TOKEN_ID
    {
        if(symbol_table.count(*$1) == 0) {
			std::stringstream ss;
			ss << "Undeclared variable: " << *$1;
			error(ss.str().c_str());
        }
        $$ = new type(symbol_table[*$1].var_type);
		delete $1;
    }
|
    expression TOKEN_ADD expression
    {
		if(*$1 != type_integer) {
			error("Left operand of '+' is not integer.");
		}
		if(*$3 != type_integer) {
			error("Right operand of '+' is not integer.");
		}
		$$ = new type(type_integer);
		delete $1;
		delete $3;
    }
|
    expression TOKEN_SUB expression
    {
		if(*$1 != type_integer) {
			error("Left operand of '-' is not integer.");
		}
		if(*$3 != type_integer) {
			error("Right operand of '-' is not integer.");
		}
		$$ = new type(type_integer);
		delete $1;
		delete $3;
    }
|
    expression TOKEN_MUL expression
    {
		if(*$1 != type_integer) {
			error("Left operand of '*' is not integer.");
		}
		if(*$3 != type_integer) {
			error("Right operand of '*' is not integer.");
		}
		$$ = new type(type_integer);
		delete $1;
		delete $3;
    }
|
    expression TOKEN_DIV expression
    {
		if(*$1 != type_integer) {
			error("Left operand of 'div' is not integer.");
		}
		if(*$3 != type_integer) {
			error("Right operand of 'div' is not integer.");
		}
		$$ = new type(type_integer);
		delete $1;
		delete $3;
    }
|
    expression TOKEN_MOD expression
    {
		if(*$1 != type_integer) {
			error("Left operand of 'mod' is not integer.");
		}
		if(*$3 != type_integer) {
			error("Right operand of 'mod' is not integer.");
		}
		$$ = new type(type_integer);
		delete $1;
		delete $3;
    }
|
    expression TOKEN_LESS expression
    {
		if(*$1 != type_integer) {
			error("Left operand of '<' is not integer.");
		}
		if(*$3 != type_integer) {
			error("Right operand of '<' is not integer.");
		}
		$$ = new type(type_boolean);
		delete $1;
		delete $3;
    }
|
    expression TOKEN_GR expression
    {
		if(*$1 != type_integer) {
			error("Left operand of '>' is not integer.");
		}
		if(*$3 != type_integer) {
			error("Right operand of '>' is not integer.");
		}
		$$ = new type(type_boolean);
		delete $1;
		delete $3;
    }
|
    expression TOKEN_EQ expression
    {
		if(*$1 != *$3) {
			error("Left and right operands of '=' have different types.");
		}
		$$ = new type(type_boolean);
		delete $1;
		delete $3;
    }
|
    expression TOKEN_AND expression
    {
		if(*$1 != type_boolean) {
			error("Left operand of 'and' is not boolean.");
		}
		if(*$3 != type_boolean) {
			error("Right operand of 'and' is not boolean.");
		}
		$$ = new type(type_boolean);
		delete $1;
		delete $3;
    }
|
    expression TOKEN_OR expression
    {
		if(*$1 != type_boolean) {
			error("Left operand of 'or' is not boolean.");
		}
		if(*$3 != type_boolean) {
			error("Right operand of 'or' is not boolean.");
		}
		$$ = new type(type_boolean);
		delete $1;
		delete $3;
    }
|
    TOKEN_NOT expression
    {
		if(*$2 != type_boolean) {
			error("Operand of 'not' is not boolean.");
		}
		$$ = new type(type_boolean);
		delete $2;
    }
|
    TOKEN_OPEN expression TOKEN_CLOSE
    {
        $$ = $2;
    }
;
